from __future__ import annotations

from io import BytesIO
from pathlib import Path
from typing import List, Literal, Optional, Tuple, Union

from PIL import Image as PilImage, ImageDraw, ImageFilter, ImageFont
from PIL.Image import Image

from .canvas import Canvas
from .font import Font
from .text import Text
from .types.common import Color


class Editor:
    """Editor class. It does all the editing operations.

    Parameters
    ----------
    _image : Union[Image, str, Editor, Canvas]
        Image or Canvas to edit.
    """

    def __init__(
        self, _image: Union[Image, str, BytesIO, Editor, Canvas, Path]
    ) -> None:
        if isinstance(_image, (str, BytesIO, Path)):
            self.image: Image = PilImage.open(_image)
        elif isinstance(_image, (Canvas, Editor)):
            self.image: Image = _image.image
        elif isinstance(_image, Image):
            self.image: Image = _image
        else:
            raise ValueError(
                "Editor requires an Image, Path, "
                "Editor or Canvas to start with"
            )

        self.image = self.image.convert("RGBA")

    @property
    def image_bytes(self) -> BytesIO:
        """Return image bytes

        Returns
        -------
        BytesIO
            Bytes from the image of Editor
        """
        _bytes = BytesIO()
        self.image.save(_bytes, "png")

        _bytes.seek(0)
        return _bytes

    def close(self):
        self.image.close()

    def resize(self, size: Tuple[float, float], crop=False) -> Editor:
        """Resize image

        Parameters
        ----------
        size : Tuple[float, float]
            New Size of image
        crop : bool, optional
            Crop the image to bypass distortion, by default False
        """
        if not crop:
            self.image = self.image.resize(size, PilImage.LANCZOS)

        else:
            width, height = self.image.size
            ideal_width, ideal_height = size

            aspect = width / height
            ideal_aspect = ideal_width / ideal_height

            if aspect > ideal_aspect:
                new_width = ideal_aspect * height
                offset = int((width - new_width) / 2)
                resize = (offset, 0, width - offset, height)
            else:
                new_height = width / ideal_aspect
                offset = int((height - new_height) / 2)
                resize = (0, offset, width, height - offset)

            self.image = self.image.crop(resize).resize(
                (ideal_width, ideal_height), PilImage.LANCZOS
            )

        return self

    def rounded_corners(self, radius: int = 10, offset: int = 2) -> Editor:
        """Make image rounded corners

        Parameters
        ----------
        radius : int, optional
            Radius of roundness, by default 10
        offset : int, optional
            Offset pixel while making rounded, by default 2
        """
        background = PilImage.new(
            "RGBA", size=self.image.size, color=(255, 255, 255, 0)
        )
        holder = PilImage.new(
            "RGBA", size=self.image.size, color=(255, 255, 255, 0)
        )
        mask = PilImage.new(
            "RGBA", size=self.image.size, color=(255, 255, 255, 0)
        )
        mask_draw = ImageDraw.Draw(mask)
        mask_draw.rounded_rectangle(
            (offset, offset)
            + (self.image.size[0] - offset, self.image.size[1] - offset),
            radius=radius,
            fill="black",
        )
        holder.paste(self.image, (0, 0))
        self.image = PilImage.composite(holder, background, mask)

        background.close()
        holder.close()
        mask.close()

        return self

    def circle_image(self) -> Editor:
        """Make image circle"""
        background = PilImage.new(
            "RGBA", size=self.image.size, color=(255, 255, 255, 0)
        )
        holder = PilImage.new(
            "RGBA", size=self.image.size, color=(255, 255, 255, 0)
        )
        mask = PilImage.new(
            "RGBA", size=self.image.size, color=(255, 255, 255, 0)
        )
        mask_draw = ImageDraw.Draw(mask)
        ellipse_size = tuple(i - 1 for i in self.image.size)
        mask_draw.ellipse((0, 0) + ellipse_size, fill="black")
        holder.paste(self.image, (0, 0))
        self.image = PilImage.composite(holder, background, mask)

        background.close()
        holder.close()
        mask.close()

        return self

    def rotate(self, deg: float = 0, expand: bool = False) -> Editor:
        """Rotate image

        Parameters
        ----------
        deg : float, optional
            Degrees to rotate, by default 0
        expand : bool, optional
            Expand while rotating, by default False
        """
        self.image = self.image.rotate(deg, expand=expand)
        return self

    def blur(
        self, mode: Literal["box", "gaussian"] = "gaussian", amount: float = 1
    ) -> Editor:
        """Blur image

        Parameters
        ----------
        mode : Literal["box", "gaussian"], optional
            Blur mode, by default "gaussian"
        amount : float, optional
            Amount of blur, by default 1
        """
        if mode == "box":
            self.image = self.image.filter(ImageFilter.BoxBlur(radius=amount))
        if mode == "gaussian":
            self.image = self.image.filter(
                ImageFilter.GaussianBlur(radius=amount)
            )

        return self

    def blend(
        self,
        image: Union[Image, Editor, Canvas],
        alpha: float = 0.0,
        on_top: bool = False,
    ) -> Editor:
        """Blend image into editor image

        Parameters
        ----------
        image : Union[Image, Editor, Canvas]
            Image to blend
        alpha : float, optional
            Alpha amount, by default 0.0
        on_top : bool, optional
            Places image on top, by default False
        """
        if isinstance(image, Editor) or isinstance(image, Canvas):
            image = image.image

        if image.size != self.image.size:
            image = Editor(image).resize(self.image.size, crop=True).image

        if on_top:
            self.image = PilImage.blend(self.image, image, alpha=alpha)
        else:
            self.image = PilImage.blend(image, self.image, alpha=alpha)

        return self

    def paste(
        self,
        image: Union[Image, Editor, Canvas],
        position: Tuple[int, int],
    ) -> Editor:
        """Paste image into editor

        Parameters
        ----------
        image : Union[Image, Editor, Canvas]
            Image to paste
        position : Tuple[int, int]
            Position to paste
        """
        blank = PilImage.new(
            "RGBA", size=self.image.size, color=(255, 255, 255, 0)
        )

        if isinstance(image, Editor) or isinstance(image, Canvas):
            image = image.image

        blank.paste(image, position)
        self.image = PilImage.alpha_composite(self.image, blank)

        blank.close()

        return self

    def text(
        self,
        position: Tuple[float, float],
        text: str,
        font: Optional[Union[ImageFont.FreeTypeFont, Font]] = None,
        color: Color = "black",
        align: Literal["left", "center", "right"] = "left",
        stroke_width: Optional[int] = None,
        stroke_fill: Color = "black",
    ) -> Editor:
        """Draw text into image

        Parameters
        ----------
        position : Tuple[float, float]
            Position to draw text.
        text : str
            Text to draw
        font : Union[ImageFont.FreeTypeFont, Font], optional
            Font used for text, by default None
        color : Color, optional
            Color of the font, by default "black"
        align : Literal["left", "center", "right"], optional
            Align text, by default "left"
        stroke_width : int, optional
            Whether there should be any stroke. Defaults to
            None. It represents the width of the said stroke.
        stroke_fill : Color, optional
            Color of the stroke, if any stroke is applied to the
            text. Defaults to "black"
        """
        if isinstance(font, Font):
            font = font.font

        anchors = {"left": "lt", "center": "mt", "right": "rt"}

        draw = ImageDraw.Draw(self.image)

        if stroke_width:
            draw.text(
                position,
                text,
                color,
                font=font,
                anchor=anchors[align],
                stroke_width=stroke_width,
                stroke_fill=stroke_fill,
            )
        else:
            draw.text(position, text, color, font=font, anchor=anchors[align])

        return self

    def multi_text(
        self,
        position: Tuple[float, float],
        texts: List[Text],
        space_separated: bool = True,
        align: Literal["left", "center", "right"] = "left",
    ) -> Editor:
        """Draw multicolor text

        Parameters
        ----------
        position : Tuple[float, float]
            Position to draw text
        texts : List[Text]
            List of texts
        space_separated : bool, optional
            Separate texts with space, by default True
        align : Literal["left", "center", "right"], optional
            Align texts, by default "left"
        """
        draw = ImageDraw.Draw(self.image)

        if align == "left":
            position = position

        if align == "right":
            total_width = 0

            for t in texts:
                total_width += t.font.getlength(t.text)

            position = (int(position[0] - total_width), int(position[1]))

        if align == "center":
            total_width = 0

            for t in texts:
                total_width += t.font.getlength(t.text)

            position = (int(position[0] - (total_width / 2)), int(position[1]))

        for text in texts:
            sentence = text.text
            font = text.font
            color = text.color

            if space_separated:
                width = font.getlength(sentence + " ")
            else:
                width = font.getlength(sentence)

            draw.text(position, sentence, color, font=font, anchor="lm")
            position = (int(position[0] + width), int(position[1]))

        return self

    def rectangle(
        self,
        position: Tuple[float, float],
        width: float,
        height: float,
        fill: Optional[Color] = None,
        color: Optional[Color] = None,
        outline: Optional[Color] = None,
        stroke_width: float = 1,
        radius: int = 0,
    ) -> Editor:
        """Draw rectangle into image

        Parameters
        ----------
        position : Tuple[float, float]
            Position to draw rectangle
        width : float
            Width of rectangle
        height : float
            Height of rectangle
        fill : Color, optional
            Fill color, by default None
        color : Color, optional
            Alias of fill, by default None
        outline : Color, optional
            Outline color, by default None
        stroke_width : float, optional
            Stroke width, by default 1
        radius : int, optional
            Radius of rectangle, by default 0
        """
        draw = ImageDraw.Draw(self.image)

        to_width = width + position[0]
        to_height = height + position[1]

        if color:
            fill = color

        if radius <= 0:
            draw.rectangle(
                position + (to_width, to_height),
                fill=fill,
                outline=outline,
                width=stroke_width,
            )
        else:
            draw.rounded_rectangle(
                position + (to_width, to_height),
                radius=radius,
                fill=fill,
                outline=outline,
                width=stroke_width,
            )

        return self

    def bar(
        self,
        position: Tuple[int, int],
        max_width: Union[int, float],
        height: Union[int, float],
        percentage: int = 1,
        fill: Optional[Color] = None,
        color: Optional[Color] = None,
        outline: Optional[Color] = None,
        stroke_width: float = 1,
        radius: int = 0,
    ) -> Editor:
        """Draw a progress bar

        Parameters
        ----------
        position : Tuple[int, int]
            Position to draw bar
        max_width : Union[int, float]
            Max width of the bar
        height : Union[int, float]
            Height of the bar
        percentage : int, optional
            Percentage to fill of the bar, by default 1
        fill : Color, optional
            Fill color, by default None
        color : Color, optional
            Alias of fill, by default None
        outline : Color, optional
            Outline color, by default None
        stroke_width : float, optional
            Stroke width, by default 1
        radius : int, optional
            Radius of the bar, by default 0
        """
        if percentage == 0:
            return self

        if color:
            fill = color

        bg = PilImage.new("RGBA", (int(max_width), int(height)), (0, 0, 0, 0))
        main = PilImage.new(
            "RGBA", (int(max_width), int(height)), (0, 0, 0, 0)
        )
        mask = PilImage.new("L", (int(max_width), int(height)), 0)
        main_draw = ImageDraw.Draw(main)

        if percentage > 100 or percentage < 0:
            raise ValueError("Percentage must be between 1 and 100")

        bar_width = int((max_width / 100) * percentage)

        if radius <= 0:
            main_draw.rectangle(
                (0, 0) + (bar_width, height),
                fill=fill,
                outline=outline,
                width=stroke_width,
            )
        else:
            main_draw.rounded_rectangle(
                (0, 0) + (bar_width, height),
                radius=radius,
                fill=fill,
                outline=outline,
                width=stroke_width,
            )

        mask_draw = ImageDraw.Draw(mask)
        mask_draw.rounded_rectangle(
            (0, 0) + (max_width, height),
            radius=radius,
            fill=255,
            outline=255,
            width=stroke_width,
        )

        final = PilImage.composite(main, bg, mask)
        self.paste(final, position)

        main.close()
        final.close()
        bg.close()
        mask.close()

        return self

    def rounded_bar(
        self,
        position: Tuple[float, float],
        width: Union[int, float],
        height: Union[int, float],
        percentage: float,
        fill: Optional[Color] = None,
        color: Optional[Color] = None,
        stroke_width: float = 1,
    ) -> Editor:
        """Draw a rounded bar

        Parameters
        ----------
        position : Tuple[float, float]
            Position to draw rounded bar
        width : Union[int, float]
            Width of the bar
        height : Union[int, float]
            Height of the bar
        percentage : float
            Percentage to fill.
        fill : Color, optional
            Fill color, by default None
        color : Color, optional
            Alias of color, by default None
        stroke_width : float, optional
            Stroke width, by default 1
        """
        draw = ImageDraw.Draw(self.image)

        if color:
            fill = color

        start = -90
        end = (percentage * 3.6) - 90

        draw.arc(
            position + (position[0] + width, position[1] + height),
            start,
            end,
            fill,
            width=stroke_width,
        )

        return self

    def ellipse(
        self,
        position: Tuple[float, float],
        width: float,
        height: float,
        fill: Optional[Color] = None,
        color: Optional[Color] = None,
        outline: Optional[Color] = None,
        stroke_width: float = 1,
    ) -> Editor:
        """Draw an ellipse

        Parameters
        ----------
        position : Tuple[float, float]
            Position to draw ellipse
        width : float
            Width of ellipse
        height : float
            Height of ellipse
        fill : Color, optional
            Fill color, by default None
        color : Color, optional
            Alias of fill, by default None
        outline : Color, optional
            Outline color, by default None
        stroke_width : float, optional
            Stroke width, by default 1
        """
        draw = ImageDraw.Draw(self.image)
        to_width = width + position[0]
        to_height = height + position[1]

        if color:
            fill = color

        draw.ellipse(
            position + (to_width, to_height),
            outline=outline,
            fill=fill,
            width=stroke_width,
        )

        return self

    def polygon(
        self,
        coordinates: list,
        fill: Optional[Color] = None,
        color: Optional[Color] = None,
        outline: Optional[Color] = None,
    ) -> Editor:
        """Draw a polygon

        Parameters
        ----------
        coordinates : list
            Coordinates to draw
        fill : Color, optional
            Fill color, by default None
        color : Color, optional
            Alias of fill, by default None
        outline : Color, optional
            Outline color, by default None
        """
        if color:
            fill = color

        draw = ImageDraw.Draw(self.image)
        draw.polygon(coordinates, fill=fill, outline=outline)

        return self

    def arc(
        self,
        position: Tuple[float, float],
        width: float,
        height: float,
        start: float,
        rotation: float,
        fill: Optional[Color] = None,
        color: Optional[Color] = None,
        stroke_width: float = 1,
    ) -> Editor:
        """Draw arc

        Parameters
        ----------
        position : Tuple[float, float]
            Position to draw arc
        width : float
            Width or arc
        height : float
            Height of arch
        start : float
            Start position of arch
        rotation : float
            Rotation in degree
        fill : Color, optional
            Fill color, by default None
        color : Color, optional
            Alias of fill, by default None
        stroke_width : float, optional
            Stroke width, by default 1
        """
        draw = ImageDraw.Draw(self.image)

        start = start - 90
        end = rotation - 90

        if color:
            fill = color

        draw.arc(
            position + (position[0] + width, position[1] + height),
            start,
            end,
            fill,
            width=stroke_width,
        )

        return self

    def show(self):
        """Show the image."""
        self.image.show()

    def save(self, fp, file_format: Optional[str] = None, **params):
        """Save the image

        Parameters
        ----------
        fp : str
            File path
        file_format : str, optional
            File format, by default None
        """
        self.image.save(fp, file_format, **params)



from io import BytesIO
from pathlib import Path
from typing import List, Tuple, Union

try:
    from typing import Literal, NotRequired, TypedDict
except ImportError:
    from typing_extensions import TypedDict, NotRequired, Literal

from PIL.Image import Image
from PIL.ImageFont import FreeTypeFont

from ..canvas import Canvas
from ..editor import Editor
from ..font import Font
from ..text import Text


class ComponentKwargs(TypedDict):
    size: NotRequired[Tuple[float, float]]
    position: NotRequired[Tuple[float, float]]
    crop: NotRequired[bool]
    radius: NotRequired[int]
    offset: NotRequired[int]
    deg: NotRequired[float]
    expand: NotRequired[bool]
    mode: NotRequired[Literal["box", "gaussian"]]
    amount: NotRequired[float]
    image: NotRequired[Union[Image, Editor, Canvas, BytesIO, Path, bytes]]
    alpha: NotRequired[float]
    on_top: NotRequired[bool]
    text: NotRequired[str]
    font: NotRequired[Union[FreeTypeFont, Font]]
    align: NotRequired[Literal["left", "center", "right"]]
    color: NotRequired[
        Union[int, str, Tuple[int, int, int], Tuple[int, int, int, int]]
    ]
    fill: NotRequired[
        Union[int, str, Tuple[int, int, int], Tuple[int, int, int, int]]
    ]
    space_separated: NotRequired[bool]
    texts: NotRequired[List[Text]]
    width: NotRequired[float]
    height: NotRequired[float]
    stoke_width: NotRequired[float]
    outline: NotRequired[float]
    max_width: NotRequired[float]
    percent: NotRequired[int]
    start: NotRequired[float]
    rotation: NotRequired[int]



import asyncio
import functools
from io import BytesIO
from typing import Optional, Union

import aiohttp
import requests
from aiocache import cached
from PIL import Image
from PIL.GifImagePlugin import GifImageFile


async def run_in_executor(func, **kwargs):
    """Run function in executor

    Parameters
    ----------
    func : func
        Function to run
    """
    func = functools.partial(func, **kwargs)
    data = await asyncio.get_event_loop().run_in_executor(None, func)
    return data


def load_image(
    link: str, raw: bool = False
) -> Union[Image.Image, GifImageFile]:
    """Load image from link

    Parameters
    ----------
    link : str
        Image link
    raw: bool
        if you want the raw image without any conversion

    Returns
    -------
    PIL.Image.Image
        Image from the provided link (if any)
    """
    _bytes = BytesIO(requests.get(link).content)
    image = Image.open(_bytes)
    if not raw:
        image = image.convert("RGBA")

    return image


@cached(ttl=60 * 60 * 24)
async def load_image_async(
    link: str,
    session: Optional[aiohttp.ClientSession] = None,
    raw: bool = False,
) -> Union[Image.Image, GifImageFile]:
    """Load image from link (async)

    Parameters
    ----------
    link : str
        Image from the provided link (if any)
    session: aiohttp.ClientSession
        clientSession for making requests, defaults to None
    raw: bool
        if you want the raw image without any conversion

    Returns
    -------
    PIL.Image.Image
        Image link
    """
    if isinstance(session, aiohttp.ClientSession):
        async with session.get(link) as response:  # type: ignore
            data = await response.read()
    else:
        async with aiohttp.ClientSession() as session:
            async with session.get(link) as response:
                data = await response.read()

    _bytes = BytesIO(data)
    image = Image.open(_bytes)
    if not raw:
        image = image.convert("RGBA")

    return image



from typing import Tuple, Union

from PIL import ImageFont

from .font import Font


class Text:
    """Text class

    Parameters
    ----------
    text : str
        Text
    font : ImageFont.FreeTypeFont
        Font for text
    color : Color, optional
        Font color, by default "black"
    """

    def __init__(
        self,
        text: str,
        font: Union[ImageFont.FreeTypeFont, Font],
        color: Union[
            int, str, Tuple[int, int, int], Tuple[int, int, int, int]
        ] = "black",
    ) -> None:
        self.text = text
        self.color = color

        if isinstance(font, Font):
            self.font = font.font
        else:
            self.font = font

    def getsize(self):
        bbox = self.font.getbbox(self.text)
        return bbox[2], bbox[3]



from typing import Optional, Tuple

from PIL import Image

from .types.common import Color


class Canvas:
    """Canvas class

    Parameters
    ----------
    size : Tuple[int, int], optional
        Size of image, by default None
    width : int, optional
        Width of image, by default None
    height : int, optional
        Height of image, by default None
    color : Color, optional
        Color of image, by default None

    Raises
    ------
    ValueError
        When either ``size`` or ``width and height`` is not a provided
    """

    def __init__(
        self,
        size: Optional[Tuple[int, int]] = None,
        width: int = 0,
        height: int = 0,
        color: Color = 0,
    ) -> None:
        if not (size or (width and height)):
            raise ValueError("size, width, and height cannot all be None")

        elif not size:
            size = (width, height)

        self.size = size
        self.color = color

        self.image = Image.new("RGBA", size, color=color)



from ._version import __version__, version_info
from .aio_editor import AioEditor
from .canvas import Canvas
from .editor import Editor
from .font import Font
from .gif_editor import GifEditor
from .text import Text
from .utils import load_image, load_image_async, run_in_executor
from .workspace import Workspace

__all__ = [
    "__version__",
    "version_info",
    "Canvas",
    "Editor",
    "GifEditor",
    "AioEditor",
    "Workspace",
    "Font",
    "Text",
    "load_image",
    "load_image_async",
    "run_in_executor",
]



import os
from functools import lru_cache

from PIL import ImageFont
from typing_extensions import Literal

fonts_directory = os.path.join(os.path.dirname(__file__), "fonts")


fonts_path = {
    "caveat": {
        "regular": os.path.join(fonts_directory, "caveat", "caveat.ttf"),
        "bold": os.path.join(fonts_directory, "caveat", "caveat.ttf"),
        "italic": os.path.join(fonts_directory, "caveat", "caveat.ttf"),
        "light": os.path.join(fonts_directory, "caveat", "caveat.ttf"),
    },
    "montserrat": {
        "regular": os.path.join(
            fonts_directory, "montserrat", "montserrat_regular.ttf"
        ),
        "bold": os.path.join(
            fonts_directory, "montserrat", "montserrat_bold.ttf"
        ),
        "italic": os.path.join(
            fonts_directory, "montserrat", "montserrat_italic.ttf"
        ),
        "light": os.path.join(
            fonts_directory, "montserrat", "montserrat_light.ttf"
        ),
    },
    "poppins": {
        "regular": os.path.join(
            fonts_directory, "poppins", "poppins_regular.ttf"
        ),
        "bold": os.path.join(fonts_directory, "poppins", "poppins_bold.ttf"),
        "italic": os.path.join(
            fonts_directory, "poppins", "poppins_italic.ttf"
        ),
        "light": os.path.join(fonts_directory, "poppins", "poppins_light.ttf"),
    },
}


class Font:
    """Font class

    Parameters
    ----------
    path : str
        Path of font
    size : int, optional
        Size of font, by default 10
    """

    def __init__(self, path: str, size: int = 10, **kwargs) -> None:
        self.font = ImageFont.truetype(path, size=size, **kwargs)

    def getsize(self, text: str):
        bbox = self.font.getbbox(text)
        return bbox[2], bbox[3]

    @staticmethod
    @lru_cache(32)
    def poppins(
        variant: Literal["regular", "bold", "italic", "light"] = "regular",
        size: int = 10,
    ):
        """Poppins font

        Parameters
        ----------
        variant : Literal["regular", "bold", "italic", "light"], optional
            Font variant, by default "regular"
        size : int, optional
            Font size, by default 10
        """
        return ImageFont.truetype(fonts_path["poppins"][variant], size=size)

    @staticmethod
    @lru_cache(32)
    def caveat(
        variant: Literal["regular", "bold", "italic", "light"] = "regular",
        size: int = 10,
    ):
        """Caveat font

        Parameters
        ----------
        variant : Literal["regular", "bold", "italic", "light"], optional
            Font variant, by default "regular"
        size : int, optional
            Font size, by default 10
        """
        return ImageFont.truetype(fonts_path["caveat"][variant], size=size)

    @staticmethod
    @lru_cache(32)
    def montserrat(
        variant: Literal["regular", "bold", "italic", "light"] = "regular",
        size: int = 10,
    ):
        """Montserrat font

        Parameters
        ----------
        variant : Literal["regular", "bold", "italic", "light"], optional
            Font variant, by default "regular"
        size : int, optional
            Font size, by default 10
        """
        return ImageFont.truetype(fonts_path["montserrat"][variant], size=size)



import random
import string
from typing import Any, Callable, Dict, Optional, Tuple, Union

from .editor import Canvas, Editor
from .types.common import Color
from .types.workspace import ComponentKwargs


class Workspace:
    """Workspace class for working with layers and components"""

    def __init__(self, size: Tuple[int, int]) -> None:
        self.size = size
        self.layers: dict = dict()
        self.working_layer = None

    def create_layer(self, name: str, background: Color = (0, 0, 0, 0)):
        """Creates a layer

        Parameters
        ----------
        name : str
            name of the layer
        background: Color
            background color of the layer
        """
        self.layers[name] = {
            "metadata": {
                "background": background,
            },
            "components": dict(),
        }

    def remove_layer(self, name: str):
        """Removes a layer

        Parameters
        ----------
        name : str
            name of the layer

        Raises
        ------
        ValueError
            if the layer is not available in the workspace
        """
        try:
            self.layers.pop(name)
        except KeyError:
            raise ValueError("Invalid layer name")

    def update_layer(
        self,
        layer_name: str,
        new_layer_name: Optional[str] = None,
        background: Optional[Color] = None,
    ):
        """Creates a layer

        Parameters
        ----------
        layer_name : str
            name of the layer
        new_layer_name: str, Optional
            updated name of the layer, defaults to None
        background: Color
            background color of the layer, defaults to None

        Raises
        ------
        ValueError
            if the layer is not available in the workspace
        """
        if layer_name not in self.layers:
            raise ValueError("Invalid layer name")

        if background:
            self.layers[layer_name]["metadata"]["background"] = background

        if new_layer_name:
            self.layers[new_layer_name] = self.layers.pop(layer_name)

    def set_working_layer(self, name: str):
        """Sets a layer as working layer

        Parameters
        ----------
        name : str
            name of the layer

        Raises
        ------
        ValueError
            if the layer is not available in the workspace
        """
        if name not in self.layers:
            raise ValueError("Invalid layer name")

        self.working_layer = name

    def __get_random_identifier(self) -> str:
        return "".join(random.choices(string.ascii_letters, k=5))

    def add_component(
        self,
        *,
        layer_name: Optional[str] = None,
        identifier: Optional[str] = None,
        func: Union[Callable, str],
        options: ComponentKwargs
    ):
        """Add component to a layer

        Parameters
        ----------
        layer_name : str
            name of the layer
        identifier : str
            unique name for a component
        func : Union[Callable, str]
            the function or function name from editor class
        options : ComponentKwargs
            keyword arguments for the func

        Raises
        ------
        ValueError
            if the layer is not available in the workspace
        """
        if not self.working_layer and not layer_name:
            raise ValueError(
                "Either specify layer name or set a working layer"
            )

        layer_name = layer_name or self.working_layer

        if layer_name not in self.layers:
            raise ValueError("Invalid layer name")

        func_name = func.__name__ if isinstance(func, Callable) else func
        identifier_name = (
            identifier if identifier else self.__get_random_identifier()
        )

        self.layers[layer_name]["components"][identifier_name] = {
            "func_name": func_name,
            "options": options,
        }

    def remove_component(
        self, *, layer_name: Optional[str] = None, identifier: str
    ):
        """Remove component from a layer

        Parameters
        ----------
        layer_name : str
            name of the layer
        identifier : str
            unique name for a component

        Raises
        ------
        ValueError
            if the layer is not available in the workspace
        """

        if not self.working_layer and not layer_name:
            raise ValueError(
                "Either specify layer name or set a working layer"
            )

        layer_name = layer_name or self.working_layer

        try:
            self.layers[layer_name]["components"].pop(identifier)
        except KeyError:
            raise ValueError("Invalid layer name or identifier")

    def update_component(
        self,
        *,
        layer_name: Optional[str] = None,
        identifier: str,
        options: ComponentKwargs
    ):
        """Update component of a layer

        Parameters
        ----------
        layer_name : str
            name of the layer
        identifier : str
            unique name for a component
        options : ComponentKwargs
            modified options

        Raises
        ------
        ValueError
            if the layer is not available in the workspace
        """
        if not self.working_layer and not layer_name:
            raise ValueError(
                "Either specify layer name or set a working layer"
            )

        layer_name = layer_name or self.working_layer

        if layer_name not in self.layers:
            raise ValueError("Invalid layer name")

        self.layers[layer_name]["components"][identifier]["options"].update(
            options
        )

    def __create_editor_layer(
        self, size: Tuple[int, int], metadata: Dict[str, Any]
    ):
        return Editor(Canvas(size, color=metadata["background"]))

    def generate_image(self) -> Editor:
        """Generates image from the layers

        Returns
        -------
        Editor
            The editor instance
        """
        editor = Editor(Canvas(self.size, color=(0, 0, 0, 0)))

        for layer in self.layers.values():
            _layer = self.__create_editor_layer(self.size, layer["metadata"])

            for config in layer["components"].values():
                func_name = config["func_name"]
                options = config["options"]

                _func = getattr(_layer, func_name)

                if _func:
                    _func(**options)

            editor.paste(_layer, position=(0, 0))

        return editor



from collections import namedtuple

__version__ = "0.3.8"

VersionInfo = namedtuple("VersionInfo", "major minor macro release")

version_info = VersionInfo(*map(int, __version__.split(".")), release="stable")



from __future__ import annotations

from io import BytesIO
from pathlib import Path
from typing import List, Tuple, Union

from PIL import Image as PilImage, ImageSequence
from PIL.GifImagePlugin import GifImageFile

from .editor import Editor


class GifEditor:
    def __init__(self, image: Union[str, BytesIO, Path, GifImageFile]):
        if isinstance(image, (str, BytesIO, Path)):
            self.image = PilImage.open(image)
        if isinstance(image, GifImageFile):
            self.image = image

        self.original_frames = ImageSequence.Iterator(self.image)
        self.frames: List[Editor] = list(
            map(lambda x: Editor(x), self.original_frames)
        )
        self.size: Tuple[int, int] = self.image.size

    def __getattr__(self, name):
        def wrapper(*args, **kwargs):
            for frame in self.frames:
                getattr(frame, name)(*args, **kwargs)

        return wrapper

    @property
    def image_bytes(self) -> BytesIO:
        """Return image bytes

        Returns
        -------
        BytesIO
            Bytes from the image of Editor
        """
        _bytes = BytesIO()
        images = list(map(lambda e: e.image, self.frames))
        images[0].save(_bytes, "GIF", save_all=True, append_images=images[1:])

        _bytes.seek(0)
        return _bytes

    def save(self, fp, **kwargs):
        """Save the image

        Parameters
        ----------
        fp : str
            File path
        """
        images = list(map(lambda e: e.image, self.frames))
        images[0].save(
            fp, "GIF", save_all=True, append_images=images[1:], **kwargs
        )



from __future__ import annotations

import asyncio
from dataclasses import dataclass, field
from functools import partial
from io import BytesIO
from pathlib import Path
from typing import Any, Dict, List, Sequence, Union

from PIL.Image import Image

from .canvas import Canvas
from .editor import Editor


@dataclass
class Instruction:
    name: str
    args: Sequence = field(default_factory=lambda: [])
    kwargs: Dict[str, Any] = field(default_factory=lambda: {})


class AioEditor:
    def __init__(
        self, _image: Union[Image, str, BytesIO, Editor, Canvas, Path]
    ) -> None:
        self.image = _image
        self.instructions: List[Instruction] = []

    def __getattr__(self, name):
        if hasattr(Editor, name):

            def handler(*args, **kwargs):
                self.instructions.append(
                    Instruction(name=name, args=args, kwargs=kwargs)
                )

            return handler
        raise AttributeError(f"'{name}' is not available in Editor")

    async def execute(self):
        editor = Editor(self.image)
        for ins in self.instructions:
            func = partial(
                editor.__getattribute__(ins.name), *ins.args, **ins.kwargs
            )
            await asyncio.get_event_loop().run_in_executor(None, func)

        return editor
